"""Octomachinery event dispatchers collection proxy."""

# pylint: disable=unused-import
from ...routing.routers import (  # noqa: F401
    ConcurrentRouter, GidgetHubRouterBase, NonBlockingConcurrentRouter,
)



"""GitHub webhooks routing proxy."""

# pylint: disable=unused-import
from ...routing.default_router import (  # noqa: F401
    WEBHOOK_EVENTS_ROUTER, dispatch_event, process_event,
    process_event_actions,
)



"""Tests for CLI utility functions."""

from io import StringIO
from textwrap import dedent
from uuid import uuid1, uuid4

import multidict

import pytest

from octomachinery.github.utils.event_utils import (
    _probe_json, _transform_http_headers_list_to_multidict,
    augment_http_headers, make_http_headers_from_event,
    parse_event_stub_from_fd, validate_http_headers,
)


UNCHANGED_UUID4_STR = str(uuid4())


@pytest.mark.parametrize(
    'vcr_contents, vcr_headers',
    (
        pytest.param(
            dedent(
                """
                ---
                # This needs to be a sequence of mappings;
                # it cannot be just a mapping because headers
                # may occur multiple times
                - Content-Type: application/json
                - X-GitHub-Delivery: 2791443c-641a-40fa-836d-031a26f0d45f
                - X-GitHub-Event: ping
                ---
                {
                  "hook": {"app_id": 0},
                  "hook_id": 0,
                  "zen": "Hey zen!"
                }
                """,
            ),
            multidict.CIMultiDict({
                'Content-Type': 'application/json',
                'X-GitHub-Delivery': '2791443c-641a-40fa-836d-031a26f0d45f',
                'X-GitHub-Event': 'ping',
            }),
            id='YAML',
        ),
        pytest.param(
            # pylint: disable=line-too-long
            """
            [{"Content-Type": "application/json"}, {"X-GitHub-Delivery": "2791443c-641a-40fa-836d-031a26f0d45f"}, {"X-GitHub-Event": "ping"}]
            {"hook": {"app_id": 0}, "hook_id": 0, "zen": "Hey zen!"}
            """.strip(),  # noqa: E501
            multidict.CIMultiDict({
                'Content-Type': 'application/json',
                'X-GitHub-Delivery': '2791443c-641a-40fa-836d-031a26f0d45f',
                'X-GitHub-Event': 'ping',
            }),
            id='JSONL',
        ),
        pytest.param(
            '{\n"hook": {\n"app_id": 0}, "hook_id": 0, "zen": "Hey zen!"}',
            multidict.CIMultiDict(),
            id='JSON',
        ),
        pytest.param(
            """
            {"hook": {"app_id": 0}, "hook_id": 0, "zen": "Hey zen!"}
            {
            """.strip(),
            multidict.CIMultiDict(),
            id='JSONL with the broken 2nd line',
        ),
    ),
)
def test_parse_event_stub_from_fd(vcr_contents, vcr_headers):
    """Check that all of YAML, JSONL and JSON VCR modes are loadable."""
    vcr_event = {
        'hook': {
            'app_id': 0,
        },
        'hook_id': 0,
        'zen': 'Hey zen!',
    }
    with StringIO(vcr_contents) as event_file_fd:
        actual_parsed_vcr = parse_event_stub_from_fd(event_file_fd)

    expected_parsed_vcr = vcr_headers, vcr_event
    assert actual_parsed_vcr == expected_parsed_vcr


@pytest.mark.parametrize(
    'vcr_contents',
    (
        pytest.param(
            dedent(
                """
                ---
                - X-GitHub-Event: ping
                ---
                {}
                ---
                - extra document
                """,
            ),
            id='broken 3-document YAML (must be 2 documents)',
        ),
        pytest.param(
            dedent(
                """
                ---
                """,
            ),
            id='broken empty-document YAML',
        ),
        pytest.param(
            """
            [{"X-GitHub-Event": "ping"}]
            {"hook": {"app_id": 0}, "hook_id": 0, "zen": "Hey zen!"}
            {}
            """,
            id='broken 3-line JSONL (must be 2 lines)',
        ),
        pytest.param(
            '1{"hook": {"app_id": 0}, "hook_id": 0, "zen": "Hey zen!"}',
            id='broken JSON syntax (leading garbage)',
        ),
    ),
)
def test_parse_event_stub_from_fd__invalid(vcr_contents):
    """Verify that feeding unconventional VCRs raises ValueError."""
    expected_error_message = (
        r'^The input event VCR file has invalid structure\. '
        r'It must be either of YAML, JSONL or JSON\.$'
    )
    with StringIO(vcr_contents) as file_descr, pytest.raises(
            ValueError,
            match=expected_error_message,
    ):
        parse_event_stub_from_fd(file_descr)


@pytest.mark.parametrize(
    'http_headers',
    (
        pytest.param(
            multidict.CIMultiDict({
                'Content-Type': 'application/json',
                'User-Agent': 'GitHub-Hookshot/dict-test',
                'X-GitHub-Delivery': str(uuid4()),
                'X-GitHub-Event': 'issue',
            }),
            id='multidict',
        ),
        pytest.param(
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-delivery': str(uuid4()),
                'x-github-event': 'pull_request',
            },
            id='dict',
        ),
        pytest.param(
            make_http_headers_from_event('ping'),
            id='make_http_headers_from_event constructor',
        ),
    ),
)
def test_validate_http_headers(http_headers):
    """Verify that valid headers collections don't raise exceptions."""
    assert validate_http_headers(http_headers) is None  # no exceptions raised


@pytest.mark.parametrize(
    'http_headers, error_message',
    (
        pytest.param(
            {
                'content-type': 'multipart/form-data',
            },
            r"^Content\-Type must be 'application\/json'$",
            id='Content-Type',
        ),
        pytest.param(
            {
                'content-type': 'application/json',
                'user-agent': 'Fake-GitHub-Hookshot/dict-test',
            },
            r"^User\-Agent must start with 'GitHub-Hookshot\/'$",
            id='User-Agent',
        ),
        pytest.param(
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-delivery': 'garbage',
            },
            r'^X\-GitHub\-Delivery must be of type UUID4$',
            id='garbage X-GitHub-Delivery',
        ),
        pytest.param(
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-delivery': str(uuid1()),
            },
            r'^X\-GitHub\-Delivery must be of type UUID4$',
            id='UUID1 X-GitHub-Delivery',
        ),
        pytest.param(
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-delivery': str(uuid4()),
                'x-github-event': None,
            },
            r'^X\-GitHub\-Event must be a string$',
            id='X-GitHub-Event',
        ),
    ),
)
def test_validate_http_headers__invalid(http_headers, error_message):
    """Check that invalid headers cause ValueError."""
    with pytest.raises(ValueError, match=error_message):
        validate_http_headers(http_headers)


@pytest.mark.parametrize(
    'incomplete_http_headers, expected_headers',
    (
        pytest.param(
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-delivery': UNCHANGED_UUID4_STR,
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-delivery': UNCHANGED_UUID4_STR,
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            id='unchanged',
        ),
        pytest.param(
            {
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-delivery': str(uuid4()),
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            id='Content-Type',
        ),
        pytest.param(
            {
                'content-type': 'application/json',
                'x-github-delivery': str(uuid4()),
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/fallback-value',
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            id='User-Agent',
        ),
        pytest.param(
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/dict-test',
                'x-github-event': 'pull_request',
                'x-header': 'x_value',
            },
            id='X-GitHub-Delivery',
        ),
        pytest.param(
            {
                'x-github-event': 'ping',
            },
            {
                'content-type': 'application/json',
                'user-agent': 'GitHub-Hookshot/fallback-value',
                'x-github-event': 'ping',
            },
            id='X-GitHub-Event',
        ),
    ),
)
def test_augment_http_headers(incomplete_http_headers, expected_headers):
    """Check that mandatory headers are present after augmentation."""
    augmented_headers = augment_http_headers(incomplete_http_headers)

    assert validate_http_headers(augmented_headers) is None

    original_event = incomplete_http_headers['x-github-event']
    assert augmented_headers['x-github-event'] == original_event

    for header_name, header_value in expected_headers.items():
        assert augmented_headers[header_name] == header_value


def test_make_http_headers_from_event():
    """Smoke-test fake HTTP headers constructor."""
    event_name = 'issue_comment'
    http_headers = make_http_headers_from_event(event_name)

    assert http_headers['X-GitHub-Event'] == event_name
    assert http_headers['User-Agent'].endswith('/fallback-value')
    assert validate_http_headers(http_headers) is None


def test__transform_http_headers_list_to_multidict__invalid():
    """Check the headers format validation."""
    error_message = (
        '^Headers must be a sequence of mappings '
        'because keys can repeat$'
    )
    with pytest.raises(ValueError, match=error_message):
        _transform_http_headers_list_to_multidict({})


def test__probe_json():
    """Test that JSON probe loads mappings."""
    vcr_contents = (
        '{\n"hook": {\n"app_id": 0},'
        ' "hook_id": 0, '
        '"zen": "Hey zen!"}'
    )
    vcr_headers = ()
    vcr_event = {
        'hook': {
            'app_id': 0,
        },
        'hook_id': 0,
        'zen': 'Hey zen!',
    }
    with StringIO(vcr_contents) as event_file_fd:
        actual_parsed_vcr = _probe_json(event_file_fd)

    expected_parsed_vcr = vcr_headers, vcr_event
    assert actual_parsed_vcr == expected_parsed_vcr


def test__probe_json__invalid():
    """Verify that non-mapping objects crash pure JSON probe."""
    expected_error_message = '^JSON file must only contain an object mapping$'
    with StringIO('[]') as file_descr, pytest.raises(
            ValueError,
            match=expected_error_message,
    ):
        _probe_json(file_descr)



#! /usr/bin/env python3
"""Octomachinery CLI entrypoint."""

import asyncio
import importlib
import os
import pathlib
from functools import wraps
from io import TextIOWrapper
from typing import (
    Any, Callable, Dict, FrozenSet, Iterable, Iterator, Set, Union,
)

from aiohttp.client import ClientSession

import click

# pylint: disable=relative-beyond-top-level
from ..app.config import BotAppConfig
# pylint: disable=relative-beyond-top-level
from ..app.routing.abc import OctomachineryRouterBase
# pylint: disable=relative-beyond-top-level
from ..app.routing.webhooks_dispatcher import route_github_event
# pylint: disable=relative-beyond-top-level
from ..github.api.app_client import GitHubApp
# pylint: disable=relative-beyond-top-level
from ..github.entities.action import GitHubAction
# pylint: disable=relative-beyond-top-level
# pylint: disable=relative-beyond-top-level
from ..github.models.events import GitHubEvent, GitHubWebhookEvent


@click.group()
@click.pass_context
def cli(ctx: click.Context) -> None:  # pylint: disable=unused-argument
    """Click CLI base."""


def run_async(orig_async_func: Callable[..., Any]):
    """Run the given async func in event loop."""
    @wraps(orig_async_func)
    def func_wrapper(*args: Any, **kwargs: Any):
        return asyncio.run(orig_async_func(*args, **kwargs))
    return func_wrapper


@cli.command()
@click.option('--event', '-e', prompt=False, type=str)
@click.option(
    '--payload-path', '-p',
    'event_payload',
    prompt=True,
    type=click.File(mode='r'),
)
@click.option('--token', '-t', prompt=False, type=str)
@click.option('--app', '-a', prompt=False, type=int)
@click.option('--private-key', '-P', prompt=False, type=click.File(mode='r'))
@click.option('--entrypoint-module', '-m', prompt=False, type=str)
@click.option(
    '--event-router', '-r',
    'event_routers',
    multiple=True,
    prompt=False,
    type=str,
)
@click.pass_context
@run_async
async def receive(  # pylint: disable=too-many-arguments,too-many-locals
        ctx: click.Context,
        event: str, event_payload: TextIOWrapper,
        token: str,
        app: int, private_key: TextIOWrapper,
        entrypoint_module: str,
        event_routers: Iterable[str],
) -> None:
    """Webhook event receive command."""
    app_missing_private_key = app is not None and not private_key
    if app_missing_private_key:
        ctx.fail(click.style('App requires a private key', fg='red'))

    creds_present = token or (app and private_key)
    if not creds_present:
        ctx.fail(click.style('GitHub auth credentials are missing', fg='red'))

    too_many_creds_present = token and (app or private_key)
    if too_many_creds_present:
        ctx.fail(
            click.style(
                'Please choose between a token or '
                'an app id with a private key',
                fg='red',
            ),
        )

    make_event = GitHubEvent if app is None else GitHubWebhookEvent
    try:
        gh_event = make_event.from_fixture_fd(event_payload, event=event)
    except ValueError as val_err:
        ctx.fail(click.style(str(val_err), fg='red'))

    os.environ.update(get_extra_env_vars(gh_event, token, app, private_key))

    try:
        target_routers = set(
            load_event_routers(
                entrypoint_module,
                # pylint: disable=fixme
                event_routers,  # type: ignore[arg-type]  # FIXME: typing
            ),
        )
    except AttributeError as attr_err:
        ctx.fail(
            click.style(
                f'Could not find an event router: {attr_err!s}',
                fg='red',
            ),
        )
    except ImportError as imp_err:
        ctx.fail(
            click.style(f'Could not load a module: {imp_err!s}', fg='red'),
        )

    config = BotAppConfig.from_dotenv()
    gh_app_kwargs = {'config': config.github}
    make_gh_app = GitHubApp
    if app is None:
        make_gh_app = GitHubAction
        gh_app_kwargs['metadata'] = config.action
    async with ClientSession() as http_client_session:
        github_app = make_gh_app(
            http_session=http_client_session,
            # FIXME: typing  # pylint: disable=fixme
            event_routers=target_routers or None,  # type: ignore[arg-type]
            **gh_app_kwargs,
        )
        await route_github_event(
            github_event=gh_event, github_app=github_app,
        )

    click.echo(
        click.style(
            f'Finished processing {gh_event.name!s} event!',
            fg='green',
        ),
    )


def load_event_routers(
        entrypoint_module: Union[str, None] = None,
        event_routers: Union[FrozenSet[str], Set[str]] = frozenset(),
) -> Iterator[OctomachineryRouterBase]:
    """Yield event routers from strings."""
    if entrypoint_module is not None:
        importlib.import_module(entrypoint_module)

    for router_path in event_routers:
        target_sep = ':' if ':' in router_path else '.'
        module_path, _sep, target_router = router_path.rpartition(target_sep)
        yield getattr(importlib.import_module(module_path), target_router)


def get_extra_env_vars(
        gh_event: GitHubEvent, token: str, app: int,
        private_key: TextIOWrapper,
) -> Dict[str, str]:
    """Construct additional env vars for App or Action processing."""
    env = {}

    if app is not None:
        env['OCTOMACHINERY_APP_MODE'] = 'app'

        env['GITHUB_APP_IDENTIFIER'] = str(app)
        env['GITHUB_PRIVATE_KEY'] = private_key.read()
        return env

    env['OCTOMACHINERY_APP_MODE'] = 'action'

    env['GITHUB_ACTION'] = 'Fake CLI Action'
    env['GITHUB_ACTOR'] = gh_event.payload.get('sender', {}).get('login', '')
    env['GITHUB_EVENT_NAME'] = gh_event.name
    env['GITHUB_WORKSPACE'] = str(pathlib.Path('.').resolve())
    env['GITHUB_SHA'] = gh_event.payload.get('head_commit', {}).get('id', '')
    env['GITHUB_REF'] = gh_event.payload.get('ref', '')
    env['GITHUB_REPOSITORY'] = (
        gh_event.payload.
        get('repository', {}).
        get('full_name', '')
    )
    env['GITHUB_TOKEN'] = token
    env['GITHUB_WORKFLOW'] = 'Fake CLI Workflow'
    env['GITHUB_EVENT_PATH'] = '/dev/null'

    return env


def main():
    """CLI entrypoint."""
    kwargs = {
        'prog_name': f'python3 -m {__package__}',
    } if __name__ == '__main__' else {}

    return cli(  # pylint: disable=unexpected-keyword-arg
        auto_envvar_prefix='OCTOMACHINERY_CLI_',
        obj={},
        **kwargs,
    )


__name__ == '__main__' and main()  # pylint: disable=expression-not-assigned



"""Octomachinery router base interface definitions."""

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Iterator

from gidgethub.routing import AsyncCallback


if TYPE_CHECKING:
    from ..github.models.events import GitHubEvent


__all__ = ('OctomachineryRouterBase',)


class OctomachineryRouterBase(metaclass=ABCMeta):
    """Octomachinery router base."""

    # pylint: disable=unused-argument
    @abstractmethod
    def emit_routes_for(
            self, event_name: str, event_payload: Any,
    ) -> Iterator[AsyncCallback]:
        """Emit callbacks that match given event and payload.

        :param str event_name: name of the GitHub event
        :param str event_payload: details of the GitHub event

        :yields: coroutine event handlers
        """

    # pylint: disable=unused-argument
    @abstractmethod
    async def dispatch(
            self, event: 'GitHubEvent',
            *args: Any, **kwargs: Any,
    ) -> None:
        """Invoke coroutine handler tasks for the given event."""



"""GitHub webhook events dispatching logic."""

import asyncio
import logging
import typing
from functools import wraps
from http import HTTPStatus

from aiohttp import web
from gidgethub import BadRequest, ValidationFailure
from gidgethub.sansio import validate_event as validate_webhook_payload

# pylint: disable=relative-beyond-top-level,import-error
from ...github.models.events import GidgetHubWebhookEvent
# pylint: disable=relative-beyond-top-level,import-error
from ...routing.webhooks_dispatcher import route_github_event


__all__ = ('route_github_webhook_event',)


logger = logging.getLogger(__name__)


EVENT_LOG_TMPL = (
    'Got a{}valid X-GitHub-Event=%s '
    'with X-GitHub-Delivery=%s '
    'and X-Hub-Signature=%s'
)

EVENT_INVALID_CHUNK = 'n in'
EVENT_VALID_CHUNK = ' '

EVENT_LOG_VALID_MSG = EVENT_LOG_TMPL.format(EVENT_VALID_CHUNK)
EVENT_LOG_INVALID_MSG = EVENT_LOG_TMPL.format(EVENT_INVALID_CHUNK)


async def get_trusted_http_payload(request, webhook_secret):
    """Get a verified HTTP request body from request."""
    http_req_headers = request.headers
    is_secret_provided = webhook_secret is not None
    is_payload_signed = 'x-hub-signature' in http_req_headers

    if is_payload_signed and not is_secret_provided:
        raise ValidationFailure('secret not provided')

    if not is_payload_signed and is_secret_provided:
        raise ValidationFailure('signature is missing')

    raw_http_req_body = await request.read()

    if is_payload_signed and is_secret_provided:
        validate_webhook_payload(
            payload=raw_http_req_body,
            signature=http_req_headers['x-hub-signature'],
            secret=webhook_secret,
        )

    return raw_http_req_body


async def get_event_from_request(request, webhook_secret):
    """Retrieve Event out of HTTP request if it's valid."""
    webhook_event_signature = request.headers.get(
        'X-Hub-Signature', '<MISSING>',
    )
    try:
        http_req_body = await get_trusted_http_payload(
            request, webhook_secret,
        )
    except ValidationFailure as no_signature_exc:
        logger.error(
            EVENT_LOG_INVALID_MSG,
            request.headers.get('X-GitHub-Event'),
            request.headers.get('X-GitHub-Delivery'),
            webhook_event_signature,
        )
        logger.debug(
            'Webhook HTTP query signature validation failed because: %s',
            no_signature_exc,
        )
        raise web.HTTPForbidden from no_signature_exc

    event = GidgetHubWebhookEvent.from_http_request(
        http_req_headers=request.headers,
        http_req_body=http_req_body,
    )
    logger.info(
        EVENT_LOG_VALID_MSG,
        event.name,  # pylint: disable=no-member
        event.delivery_id,
        webhook_event_signature,
    )
    return event


def validate_allowed_http_methods(*allowed_methods: str):
    """Block disallowed HTTP methods."""
    _allowed_methods: typing.Set[str]
    if not allowed_methods:
        _allowed_methods = {'POST'}
    else:
        _allowed_methods = set(allowed_methods)

    def decorator(wrapped_function):
        @wraps(wrapped_function)
        async def wrapper(request, *, github_app, webhook_secret=None):
            if request.method not in _allowed_methods:
                raise web.HTTPMethodNotAllowed(
                    method=request.method,
                    allowed_methods=_allowed_methods,
                ) from BadRequest(HTTPStatus.METHOD_NOT_ALLOWED)
            return await wrapped_function(
                request,
                github_app=github_app,
                webhook_secret=webhook_secret,
            )
        return wrapper
    return decorator


def webhook_request_to_event(wrapped_function):
    """Pass event extracted from request into the wrapped function."""
    @wraps(wrapped_function)
    async def wrapper(request, *, github_app, webhook_secret=None):
        event = await get_event_from_request(request, webhook_secret)
        return await wrapped_function(
            github_event=event, github_app=github_app,
        )
    return wrapper


@validate_allowed_http_methods('POST')
@webhook_request_to_event
async def route_github_webhook_event(*, github_event, github_app):
    """Dispatch incoming webhook events to corresponding handlers."""
    asyncio.create_task(
        route_github_event(
            github_event=github_event,
            github_app=github_app,
        ),
    )
    event_ack_msg = (
        'GitHub event received and scheduled for processing. '
        f'It is {github_event!r}'
    )
    return web.Response(text=f'OK: {event_ack_msg!s}')



"""GitHub webhook events dispatching logic."""

from __future__ import annotations

import contextlib
import logging
from typing import Any, Iterable

from anyio import get_cancelled_exc_class
from anyio import sleep as async_sleep

import sentry_sdk

# pylint: disable=relative-beyond-top-level,import-error
from ..github.api.app_client import GitHubApp
# pylint: disable=relative-beyond-top-level,import-error
from ..github.entities.action import GitHubAction
# pylint: disable=relative-beyond-top-level,import-error
from ..github.errors import GitHubActionError
# pylint: disable=relative-beyond-top-level,import-error
from ..github.models.events import GitHubEvent
# pylint: disable=relative-beyond-top-level,import-error
from ..runtime.context import RUNTIME_CONTEXT


__all__ = ('route_github_event',)


logger = logging.getLogger(__name__)


# pylint: disable=fixme
async def route_github_event(  # type: ignore[return]  # FIXME
        *,
        github_event: GitHubEvent,
        github_app: GitHubApp,
) -> Iterable[Any]:
    """Dispatch GitHub event to corresponding handlers.

    Set up ``RUNTIME_CONTEXT`` before doing that. This is so
    the concrete event handlers have access to the API client
    and flags in runtime.
    """
    is_gh_action = isinstance(github_app, GitHubAction)
    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.IS_GITHUB_ACTION = is_gh_action
    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.IS_GITHUB_APP = not is_gh_action

    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.github_app = github_app

    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.github_event = github_event

    # pylint: disable=assigning-non-slot
    RUNTIME_CONTEXT.app_installation = None
    if is_gh_action:
        # pylint: disable=assigning-non-slot
        RUNTIME_CONTEXT.app_installation_client = github_app.api_client
    else:
        with contextlib.suppress(LookupError):
            # pylint: disable=pointless-string-statement
            """Provision an installation API client if possible.

            Some events (like `ping`) are
            happening application/GitHub-wide and are not bound to
            a specific installation. The webhook payloads of such events
            don't contain any reference to an installation.
            Some events don't even refer to a GitHub App
            (e.g. `security_advisory`).
            """
            github_install = await github_app.get_installation(github_event)
            # pylint: disable=assigning-non-slot
            RUNTIME_CONTEXT.app_installation = github_install
            # pylint: disable=assigning-non-slot
            RUNTIME_CONTEXT.app_installation_client = github_install.api_client

        # Give GitHub a sec to deal w/ eventual consistency.
        # This is only needed for events that arrive over HTTP.
        # If the dispatcher is invoked from GitHub Actions,
        # by the time it's invoked the action must be already consistently
        # distributed within GitHub's systems because spawning VMs takes time
        # and actions are executed in workflows that rely on those VMs.
        await async_sleep(1)

    try:
        return await github_app.dispatch_event(github_event)
    except GitHubActionError:
        # Bypass GitHub Actions errors as they are supposed to be a
        # mechanism for communicating outcomes and are expected.
        raise
    except get_cancelled_exc_class():
        raise
    except Exception as exc:  # pylint: disable=broad-except
        # NOTE: It's probably better to wrap each event handler with
        # NOTE: try/except and call `capture_exception()` there instead.
        # NOTE: We'll also need to figure out the magic of associating
        # NOTE: breadcrumbs with event handlers.
        sentry_sdk.capture_exception(exc)

        # NOTE: Framework-wise, these exceptions are meaningless because they
        # NOTE: can be anything random that the webhook author (octomachinery
        # NOTE: end-user) forgot to handle. There's nothing we can do about
        # NOTE: them except put in the log so that the end-user would be able
        # NOTE: to properly debug their problem by inspecting the logs.
        # NOTE: P.S. This is also where we'd inject Sentry
        if isinstance(exc.__context__, get_cancelled_exc_class()):
            # The CancelledError context is irrelevant to the
            # user-defined webhook event handler workflow so we're
            # dropping it from the logs:
            exc.__context__ = None

        logger.exception(
            'An unhandled exception happened while running webhook '
            'event handlers for "%s"...',
            github_event.name,
        )
        delivery_id_msg = (
            '' if is_gh_action
            else ' (Delivery ID: '  # type: ignore[attr-defined]
            # FIXME:  # pylint: disable=fixme
            f'{github_event.delivery_id!s})'
        )
        logger.debug(
            'The payload of "%s" event%s is: %r',
            github_event.name, delivery_id_msg, github_event.payload,
        )

        if is_gh_action:
            # NOTE: In GitHub Actions env, the app is supposed to run as
            # NOTE: a foreground single event process rather than a
            # NOTE: server for multiple events. It's okay to propagate
            # NOTE: unhandled errors so that they are spit out to the
            # NOTE: console.
            raise
    except BaseException:  # SystemExit + KeyboardInterrupt + GeneratorExit
        raise



"""GitHub App runtime context helpers."""

import logging
import os


logger = logging.getLogger(__name__)


def detect_env_mode():
    """Figure out whether we're under GitHub Action environment."""
    for var_suffix in (
            'WORKFLOW',
            'ACTION', 'ACTOR',
            'REPOSITORY',
            'EVENT_NAME', 'EVENT_PATH',
            'WORKSPACE',
            'SHA', 'REF',
            'TOKEN',
    ):
        if f'GITHUB_{var_suffix}' not in os.environ:
            logger.info(
                'Detected GitHub App mode since '
                'GITHUB_%s is missing from the env',
                var_suffix,
            )
            return 'app'
    logger.info(
        'Detected GitHub Action mode since all the '
        'typical env vars are present in the env',
    )
    return 'action'
